==> v5.11.0.tar.gz: smartdata/test/test.search.ts <==
import { tap, expect } from '@push.rocks/tapbundle';
import * as smartmongo from '@push.rocks/smartmongo';
import { smartunique } from '../ts/plugins.js';

// Import the smartdata library
import * as smartdata from '../ts/index.js';
import { searchable, getSearchableFields } from '../ts/classes.doc.js';

// Set up database connection
let smartmongoInstance: smartmongo.SmartMongo;
let testDb: smartdata.SmartdataDb;

// Define a test class with searchable fields using the standard SmartDataDbDoc
@smartdata.Collection(() => testDb)
class Product extends smartdata.SmartDataDbDoc<Product, Product> {
  @smartdata.unI()
  public id: string = smartunique.shortId();

  @smartdata.svDb()
  @searchable()
  public name: string;

  @smartdata.svDb()
  @searchable()
  public description: string;

  @smartdata.svDb()
  @searchable()
  public category: string;

  @smartdata.svDb()
  public price: number;

  constructor(nameArg: string, descriptionArg: string, categoryArg: string, priceArg: number) {
    super();
    this.name = nameArg;
    this.description = descriptionArg;
    this.category = categoryArg;
    this.price = priceArg;
  }
}

tap.test('should create a test database instance', async () => {
  smartmongoInstance = await smartmongo.SmartMongo.createAndStart();
  testDb = new smartdata.SmartdataDb(await smartmongoInstance.getMongoDescriptor());
  await testDb.init();
});

tap.test('should create test products with searchable fields', async () => {
  // Create several products with different fields to search
  const products = [
    new Product('iPhone 12', 'Latest iPhone with A14 Bionic chip', 'Electronics', 999),
    new Product('MacBook Pro', 'Powerful laptop for professionals', 'Electronics', 1999),
    new Product('AirPods', 'Wireless earbuds with noise cancellation', 'Electronics', 249),
    new Product('Galaxy S21', 'Samsung flagship phone with great camera', 'Electronics', 899),
    new Product('Kindle Paperwhite', 'E-reader with built-in light', 'Books', 129),
    new Product('Harry Potter', 'Fantasy book series about wizards', 'Books', 49),
    new Product('Coffee Maker', 'Automatic drip coffee machine', 'Kitchen', 89),
    new Product('Blender', 'High-speed blender for smoothies', 'Kitchen', 129),
  ];

  // Save all products to the database
  for (const product of products) {
    await product.save();
  }

  // Verify that we can get all products
  const allProducts = await Product.getInstances({});
  expect(allProducts.length).toEqual(products.length);
  console.log(`Successfully created and saved ${allProducts.length} products`);
});

tap.test('should retrieve searchable fields for a class', async () => {
  // Use the getSearchableFields function to verify our searchable fields
  const searchableFields = getSearchableFields('Product');
  console.log('Searchable fields:', searchableFields);

  expect(searchableFields.length).toEqual(3);
  expect(searchableFields).toContain('name');
  expect(searchableFields).toContain('description');
  expect(searchableFields).toContain('category');
});

tap.test('should search products by exact field match', async () => {
  // Basic field exact match search
  const electronicsProducts = await Product.getInstances({ category: 'Electronics' });
  console.log(`Found ${electronicsProducts.length} products in Electronics category`);

  expect(electronicsProducts.length).toEqual(4);
});

tap.test('should search products by basic search method', async () => {
  // Using the basic search method with simple Lucene query
  try {
    const iPhoneResults = await Product.search('iPhone');
    console.log(`Found ${iPhoneResults.length} products matching 'iPhone' using basic search`);

    expect(iPhoneResults.length).toEqual(1);
    expect(iPhoneResults[0].name).toEqual('iPhone 12');
  } catch (error) {
    console.error('Basic search error:', error.message);
    // If basic search fails, we'll demonstrate the enhanced approach in later tests
    console.log('Will test with enhanced searchWithLucene method next');
  }
});

tap.test('should search products with search method', async () => {
  // Using the robust searchWithLucene method
    const wirelessResults = await Product.search('wireless');
    console.log(
      `Found ${wirelessResults.length} products matching 'wireless' using search`,
    );

  expect(wirelessResults.length).toEqual(1);
  expect(wirelessResults[0].name).toEqual('AirPods');
});

tap.test('should search products by category with search', async () => {
  // Using field-specific search with searchWithLucene
  const kitchenResults = await Product.search('category:Kitchen');
  console.log(`Found ${kitchenResults.length} products in Kitchen category using search`);

  expect(kitchenResults.length).toEqual(2);
  expect(kitchenResults[0].category).toEqual('Kitchen');
  expect(kitchenResults[1].category).toEqual('Kitchen');
});

tap.test('should search products with partial word matches', async () => {
  // Testing partial word matches
  const proResults = await Product.search('Pro');
  console.log(`Found ${proResults.length} products matching 'Pro'`);

  // Should match both "MacBook Pro" and "professionals" in description
  expect(proResults.length).toBeGreaterThan(0);
});

tap.test('should search across multiple searchable fields', async () => {
  // Test searching across all searchable fields
  const bookResults = await Product.search('book');
  console.log(`Found ${bookResults.length} products matching 'book' across all fields`);

  // Should match "MacBook" in name and "Books" in category
  expect(bookResults.length).toBeGreaterThan(1);
});

tap.test('should handle case insensitive searches', async () => {
  // Test case insensitivity
  const electronicsResults = await Product.search('electronics');
  const ElectronicsResults = await Product.search('Electronics');

  console.log(`Found ${electronicsResults.length} products matching lowercase 'electronics'`);
  console.log(`Found ${ElectronicsResults.length} products matching capitalized 'Electronics'`);

  // Both searches should return the same results
  expect(electronicsResults.length).toEqual(ElectronicsResults.length);
});

tap.test('should demonstrate search fallback mechanisms', async () => {
  console.log('\n====== FALLBACK MECHANISM DEMONSTRATION ======');
  console.log('If MongoDB query fails, searchWithLucene will:');
  console.log('1. Try using basic MongoDB filters');
  console.log('2. Fall back to field-specific searches');
  console.log('3. As last resort, perform in-memory filtering');
  console.log('This ensures robust search even with complex queries');
  console.log('==============================================\n');

  // Use a simpler term that should be found in descriptions
  // Avoid using "OR" operator which requires a text index
  const results = await Product.search('high');
  console.log(`Found ${results.length} products matching 'high'`);

  // "High-speed blender" contains "high"
  expect(results.length).toBeGreaterThan(0);

  // Try another fallback example that won't need $text
  const powerfulResults = await Product.search('powerful');
  console.log(`Found ${powerfulResults.length} products matching 'powerful'`);

  // "Powerful laptop for professionals" contains "powerful"
  expect(powerfulResults.length).toBeGreaterThan(0);
});

tap.test('should explain the advantages of the integrated approach', async () => {
  console.log('\n====== INTEGRATED SEARCH APPROACH BENEFITS ======');
  console.log('1. No separate class hierarchy - keeps code simple');
  console.log('2. Enhanced convertFilterForMongoDb handles MongoDB operators');
  console.log('3. Robust fallback mechanisms ensure searches always work');
  console.log('4. searchWithLucene provides powerful search capabilities');
  console.log('5. Backwards compatible with existing code');
  console.log('================================================\n');

  expect(true).toEqual(true);
});

// Additional robustness tests
tap.test('should search exact name using field:value', async () => {
  const nameResults = await Product.search('name:AirPods');
  expect(nameResults.length).toEqual(1);
  expect(nameResults[0].name).toEqual('AirPods');
});

tap.test('should throw when searching non-searchable field', async () => {
  let error: Error;
  try {
    await Product.search('price:129');
  } catch (err) {
    error = err as Error;
  }
  expect(error).toBeTruthy();
  expect(error.message).toMatch(/not searchable/);
});

tap.test('empty query should return all products', async () => {
  const allResults = await Product.search('');
  expect(allResults.length).toEqual(8);
});

tap.test('should search multi-word term across fields', async () => {
  const termResults = await Product.search('iPhone 12');
  expect(termResults.length).toEqual(1);
  expect(termResults[0].name).toEqual('iPhone 12');
});

// Additional search scenarios
tap.test('should return zero results for non-existent terms', async () => {
  const noResults = await Product.search('NonexistentTerm');
  expect(noResults.length).toEqual(0);
});

tap.test('should search products by description term "noise"', async () => {
  const noiseResults = await Product.search('noise');
  expect(noiseResults.length).toEqual(1);
  expect(noiseResults[0].name).toEqual('AirPods');
});

tap.test('should search products by description term "flagship"', async () => {
  const flagshipResults = await Product.search('flagship');
  expect(flagshipResults.length).toEqual(1);
  expect(flagshipResults[0].name).toEqual('Galaxy S21');
});

tap.test('should search numeric strings "12"', async () => {
  const twelveResults = await Product.search('12');
  expect(twelveResults.length).toEqual(1);
  expect(twelveResults[0].name).toEqual('iPhone 12');
});

tap.test('should search hyphenated terms "high-speed"', async () => {
  const hyphenResults = await Product.search('high-speed');
  expect(hyphenResults.length).toEqual(1);
  expect(hyphenResults[0].name).toEqual('Blender');
});

tap.test('should search hyphenated terms "E-reader"', async () => {
  const ereaderResults = await Product.search('E-reader');
  expect(ereaderResults.length).toEqual(1);
  expect(ereaderResults[0].name).toEqual('Kindle Paperwhite');
});

tap.test('close database connection', async () => {
  await testDb.mongoDb.dropDatabase();
  await testDb.close();
  if (smartmongoInstance) {
    await smartmongoInstance.stopAndDumpToDir(`.nogit/dbdump/test.search.ts`);
  }
  setTimeout(() => process.exit(), 2000);
});

tap.start({ throwOnError: true });

==> v5.11.4.tar.gz: smartdata/test/test.search.ts <==
import { tap, expect } from '@push.rocks/tapbundle';
import * as smartmongo from '@push.rocks/smartmongo';
import { smartunique } from '../ts/plugins.js';

// Import the smartdata library
import * as smartdata from '../ts/index.js';
import { searchable } from '../ts/classes.doc.js';

// Set up database connection
let smartmongoInstance: smartmongo.SmartMongo;
let testDb: smartdata.SmartdataDb;

// Define a test class with searchable fields using the standard SmartDataDbDoc
@smartdata.Collection(() => testDb)
class Product extends smartdata.SmartDataDbDoc<Product, Product> {
  @smartdata.unI()
  public id: string = smartunique.shortId();

  @smartdata.svDb()
  @searchable()
  public name: string;

  @smartdata.svDb()
  @searchable()
  public description: string;

  @smartdata.svDb()
  @searchable()
  public category: string;

  @smartdata.svDb()
  public price: number;

  constructor(nameArg: string, descriptionArg: string, categoryArg: string, priceArg: number) {
    super();
    this.name = nameArg;
    this.description = descriptionArg;
    this.category = categoryArg;
    this.price = priceArg;
  }
}

tap.test('should create a test database instance', async () => {
  smartmongoInstance = await smartmongo.SmartMongo.createAndStart();
  testDb = new smartdata.SmartdataDb(await smartmongoInstance.getMongoDescriptor());
  await testDb.init();
});

tap.test('should create test products with searchable fields', async () => {
  // Create several products with different fields to search
  const products = [
    new Product('iPhone 12', 'Latest iPhone with A14 Bionic chip', 'Electronics', 999),
    new Product('MacBook Pro', 'Powerful laptop for professionals', 'Electronics', 1999),
    new Product('AirPods', 'Wireless earbuds with noise cancellation', 'Electronics', 249),
    new Product('Galaxy S21', 'Samsung flagship phone with great camera', 'Electronics', 899),
    new Product('Kindle Paperwhite', 'E-reader with built-in light', 'Books', 129),
    new Product('Harry Potter', 'Fantasy book series about wizards', 'Books', 49),
    new Product('Coffee Maker', 'Automatic drip coffee machine', 'Kitchen', 89),
    new Product('Blender', 'High-speed blender for smoothies', 'Kitchen', 129),
  ];

  // Save all products to the database
  for (const product of products) {
    await product.save();
  }

  // Verify that we can get all products
  const allProducts = await Product.getInstances({});
  expect(allProducts.length).toEqual(products.length);
  console.log(`Successfully created and saved ${allProducts.length} products`);
});

tap.test('should retrieve searchable fields for a class', async () => {
  // Use the getSearchableFields function to verify our searchable fields
  const searchableFields = Product.getSearchableFields();
  console.log('Searchable fields:', searchableFields);

  expect(searchableFields.length).toEqual(3);
  expect(searchableFields).toContain('name');
  expect(searchableFields).toContain('description');
  expect(searchableFields).toContain('category');
});

tap.test('should search products by exact field match', async () => {
  // Basic field exact match search
  const electronicsProducts = await Product.getInstances({ category: 'Electronics' });
  console.log(`Found ${electronicsProducts.length} products in Electronics category`);

  expect(electronicsProducts.length).toEqual(4);
});

tap.test('should search products by basic search method', async () => {
  // Using the basic search method with simple Lucene query
  try {
    const iPhoneResults = await Product.search('iPhone');
    console.log(`Found ${iPhoneResults.length} products matching 'iPhone' using basic search`);

    expect(iPhoneResults.length).toEqual(1);
    expect(iPhoneResults[0].name).toEqual('iPhone 12');
  } catch (error) {
    console.error('Basic search error:', error.message);
    // If basic search fails, we'll demonstrate the enhanced approach in later tests
    console.log('Will test with enhanced searchWithLucene method next');
  }
});

tap.test('should search products with search method', async () => {
  // Using the robust searchWithLucene method
    const wirelessResults = await Product.search('wireless');
    console.log(
      `Found ${wirelessResults.length} products matching 'wireless' using search`,
    );

  expect(wirelessResults.length).toEqual(1);
  expect(wirelessResults[0].name).toEqual('AirPods');
});

tap.test('should search products by category with search', async () => {
  // Using field-specific search with searchWithLucene
  const kitchenResults = await Product.search('category:Kitchen');
  console.log(`Found ${kitchenResults.length} products in Kitchen category using search`);

  expect(kitchenResults.length).toEqual(2);
  expect(kitchenResults[0].category).toEqual('Kitchen');
  expect(kitchenResults[1].category).toEqual('Kitchen');
});

tap.test('should search products with partial word matches', async () => {
  // Testing partial word matches
  const proResults = await Product.search('Pro');
  console.log(`Found ${proResults.length} products matching 'Pro'`);

  // Should match both "MacBook Pro" and "professionals" in description
  expect(proResults.length).toBeGreaterThan(0);
});

tap.test('should search across multiple searchable fields', async () => {
  // Test searching across all searchable fields
  const bookResults = await Product.search('book');
  console.log(`Found ${bookResults.length} products matching 'book' across all fields`);

  // Should match "MacBook" in name and "Books" in category
  expect(bookResults.length).toBeGreaterThan(1);
});

tap.test('should handle case insensitive searches', async () => {
  // Test case insensitivity
  const electronicsResults = await Product.search('electronics');
  const ElectronicsResults = await Product.search('Electronics');

  console.log(`Found ${electronicsResults.length} products matching lowercase 'electronics'`);
  console.log(`Found ${ElectronicsResults.length} products matching capitalized 'Electronics'`);

  // Both searches should return the same results
  expect(electronicsResults.length).toEqual(ElectronicsResults.length);
});

tap.test('should demonstrate search fallback mechanisms', async () => {
  console.log('\n====== FALLBACK MECHANISM DEMONSTRATION ======');
  console.log('If MongoDB query fails, searchWithLucene will:');
  console.log('1. Try using basic MongoDB filters');
  console.log('2. Fall back to field-specific searches');
  console.log('3. As last resort, perform in-memory filtering');
  console.log('This ensures robust search even with complex queries');
  console.log('==============================================\n');

  // Use a simpler term that should be found in descriptions
  // Avoid using "OR" operator which requires a text index
  const results = await Product.search('high');
  console.log(`Found ${results.length} products matching 'high'`);

  // "High-speed blender" contains "high"
  expect(results.length).toBeGreaterThan(0);

  // Try another fallback example that won't need $text
  const powerfulResults = await Product.search('powerful');
  console.log(`Found ${powerfulResults.length} products matching 'powerful'`);

  // "Powerful laptop for professionals" contains "powerful"
  expect(powerfulResults.length).toBeGreaterThan(0);
});

tap.test('should explain the advantages of the integrated approach', async () => {
  console.log('\n====== INTEGRATED SEARCH APPROACH BENEFITS ======');
  console.log('1. No separate class hierarchy - keeps code simple');
  console.log('2. Enhanced convertFilterForMongoDb handles MongoDB operators');
  console.log('3. Robust fallback mechanisms ensure searches always work');
  console.log('4. searchWithLucene provides powerful search capabilities');
  console.log('5. Backwards compatible with existing code');
  console.log('================================================\n');

  expect(true).toEqual(true);
});

// Additional robustness tests
tap.test('should search exact name using field:value', async () => {
  const nameResults = await Product.search('name:AirPods');
  expect(nameResults.length).toEqual(1);
  expect(nameResults[0].name).toEqual('AirPods');
});

tap.test('should throw when searching non-searchable field', async () => {
  let error: Error;
  try {
    await Product.search('price:129');
  } catch (err) {
    error = err as Error;
  }
  expect(error).toBeTruthy();
  expect(error.message).toMatch(/not searchable/);
});

tap.test('empty query should return all products', async () => {
  const allResults = await Product.search('');
  expect(allResults.length).toEqual(8);
});

tap.test('should search multi-word term across fields', async () => {
  const termResults = await Product.search('iPhone 12');
  expect(termResults.length).toEqual(1);
  expect(termResults[0].name).toEqual('iPhone 12');
});

// Additional search scenarios
tap.test('should return zero results for non-existent terms', async () => {
  const noResults = await Product.search('NonexistentTerm');
  expect(noResults.length).toEqual(0);
});

tap.test('should search products by description term "noise"', async () => {
  const noiseResults = await Product.search('noise');
  expect(noiseResults.length).toEqual(1);
  expect(noiseResults[0].name).toEqual('AirPods');
});

tap.test('should search products by description term "flagship"', async () => {
  const flagshipResults = await Product.search('flagship');
  expect(flagshipResults.length).toEqual(1);
  expect(flagshipResults[0].name).toEqual('Galaxy S21');
});

tap.test('should search numeric strings "12"', async () => {
  const twelveResults = await Product.search('12');
  expect(twelveResults.length).toEqual(1);
  expect(twelveResults[0].name).toEqual('iPhone 12');
});

tap.test('should search hyphenated terms "high-speed"', async () => {
  const hyphenResults = await Product.search('high-speed');
  expect(hyphenResults.length).toEqual(1);
  expect(hyphenResults[0].name).toEqual('Blender');
});

tap.test('should search hyphenated terms "E-reader"', async () => {
  const ereaderResults = await Product.search('E-reader');
  expect(ereaderResults.length).toEqual(1);
  expect(ereaderResults[0].name).toEqual('Kindle Paperwhite');
});

// Additional robustness tests
tap.test('should return all products for empty search', async () => {
  const searchResults = await Product.search('');
  const allProducts = await Product.getInstances({});
  expect(searchResults.length).toEqual(allProducts.length);
});

tap.test('should support wildcard plain term across all fields', async () => {
  const results = await Product.search('*book*');
  const names = results.map((r) => r.name).sort();
  expect(names).toEqual(['Harry Potter', 'Kindle Paperwhite', 'MacBook Pro']);
});

tap.test('should support wildcard plain term with question mark pattern', async () => {
  const results = await Product.search('?one?');
  const names = results.map((r) => r.name).sort();
  expect(names).toEqual(['Galaxy S21', 'iPhone 12']);
});

// Close database connection
tap.test('close database connection', async () => {
  await testDb.mongoDb.dropDatabase();
  await testDb.close();
  if (smartmongoInstance) {
    await smartmongoInstance.stopAndDumpToDir(`.nogit/dbdump/test.search.ts`);
  }
  setTimeout(() => process.exit(), 2000);
});

tap.start({ throwOnError: true });
